;;; lkk-mclick.el --- double/triple/quadruple keyclick support for emacs

;;
;; For the codes here I wrote from scratch,
;; I just don't put any copying/copyright notice other than this:
;; Copyright (C) 2000 Seiichi Namba <sn@asahi-net.email.ne.jp>
;; There is of course no warranty.
;; Thu Apr 13 18:22:26 2000
;; s.n.
;;

;;
;; Thu Mar  9 04:55:13 2000
;; Split from lkk.el to enabling this program to be used in
;; stand-alone form (lkk.el unconditionally loads up this file, though).
;; To use lkk-mclick.el, see the code of lkk-fireup() below.
;;

;;
;; Example:
;;
;; `lkk-fireup' in lkk.el can be used to control lkk with double-M-SPC
;;
;; You need some extra setup in ~/.emacs:
;;
;; (setq lkk-fireup-old-command 'just-one-space) ; M-SPC default binding.
;; ;; Printing ASCII should be bound like this. (*not* [M-SPC], which won't work)
;; (global-set-key [?\M- ] 'lkk-fireup)
;; (add-hook
;;  'lkk-mclick-load-hook
;;  (lambda ()
;;    ;; My preferred binding to fireup lkk
;;    (define-key lkk-multiple-click-map [double-M-SPC] 'lkk-toggle-lkk-mode)))
;;
;; If you forgot to setup `lkk-fireup-old-command', you would lose 
;; the command to be bound to the original single keyclick.
;;

(provide 'lkk-mclick)

(defun lkk-toggle-inhibit-multiple-keyclick (arg)
  "Toggle `lkk-inhibit-multiple-keyclick', or set values according to ARG."
  (interactive "P")
  (cond
   ((and arg
	 (if (> (prefix-numeric-value arg) 0)
	     lkk-inhibit-multiple-keyclick
	   (not lkk-inhibit-multiple-keyclick))))
   (lkk-inhibit-multiple-keyclick
    (setq lkk-inhibit-multiple-keyclick nil))
   (t (setq lkk-inhibit-multiple-keyclick t)))
  (message "Multiple keyclick hack turned %s."
	   (if lkk-inhibit-multiple-keyclick "on" "off")))

(defvar lkk-multiple-click-map
  (let ((map (make-sparse-keymap)))
;;;    (define-key map [double-SPC] 'undefined)
;;;    (define-key map [triple-SPC] 'undefined)
;;;    (define-key map [quadruple-SPC] 'undefined)
;;;    (define-key map [triple-TAB] 'undefined)
;;;    (define-key map [triple-M-TAB] 'undefined)

;;; Why are these `tab' ?  Means CTRL-CTRL key ?
;;;    (define-key map [triple-C-tab] 'undefined)
;;;    (define-key map [triple-C-M-tab] 'undefined)

;;;    (define-key map [double-TAB] 'undefined)
;;;    (define-key map [triple-TAB] 'undefined)
;;;    (define-key map [quadruple-TAB] 'undefined)

;;;    (define-key map [double-M-return] 'lkk-toggle-lkk-mode) ; The dirty trick
    map)
  "Virtual, or pseudo keymap used by function `lkk-multiple-click-crank'.

In current lkk-*-mode-map, the function `lkk-multiple-click-crank' is
called only from commands `lkk-convert', `lkk-l-self-insert',
`lkk-l-TAB', and `lkk-completion-convert' bound to SPC or TAB, and thus,
if you want a multiple keyclick to work just with setting up this map,
these SPC and TAB can only be a `root' part of multiple keyclick sequence.
Here are multiple keyclicks' mnemonics which can be used to define
a key in this map:

	[double-SPC]
	[triple-SPC]
	[quadruple-SPC]
        [double-C-tab]
        [triple-C-M-tab]
	[quadruple-TAB]

These virtual key sequences can be bound to an user defined
interactive command by `define-key' function:

    \(define-key lkk-multiple-click-map [double-M-SPC] 'dabbrev-expand\)

With this definition, you can use [double-M-SPC] while lkk-j-mode is turned
on.  `lkk-multiple-click-map' can be looked up by `lookup-key' function
as normal keymap:

    \(lookup-key lkk-multiple-click-map [double-M-SPC]\)

Aside lkk, any interactive command can be bound in this map, and be
executed if the command bound to the `root' key part \(a sequence derived
by removing double-/triple/-quadruple part\) properly call the function
`lkk-multiple-click-crank'.  An example is `lkk-fireup', which is
conceived to toggle lkk-mode by [double-M-return] from global-map.")

(defvar lkk-multiple-click-threshold 180
  "Multiple keyclick detection threshold in milliseconds.")

(defvar lkk-inhibit-multiple-keyclick t
  "Defaults t. Nil enables double, triple, or quadruple SPC, or TAB.
The multiple-click key should be bound to user defined interactive commands in
keymap `lkk-multiple-click-map'.
See also `lkk-multiple-click-map', `lkk-multiple-click-crank'.")

;; This can be defsubst -- just a subroutine called in a recursive function.
(defsubst lkk-multiple-click-execute (hits kdsc)
  "After good enough test in field, this func should be in inline."
  ;; Allow `double-' to `quadruple-' clicks only, Also support some
  (let (pseudokey fun)
    (setq pseudokey
	  (vector
	   (cond ((= hits 2) (make-symbol (concat "double-" kdsc)))
		 ((= hits 3) (make-symbol (concat "triple-" kdsc)))
		 ((= hits 4) (make-symbol (concat "quadruple-" kdsc)))))
	  fun (lookup-key lkk-multiple-click-map pseudokey))
    (if (and fun (null (eq fun 'undefined)))
	;;(funcall fun)		; NO, should be
	(call-interactively fun)
      nil)))

;;
;; As a matter of course, lkk-multiple-click-crank() can't be defsubst.
;; (won't byte-compile) -- but sometime it ain't a matter of course:
;; I lost two hours with it.
;;

(defun lkk-multiple-click-crank (&optional hits kdsc continue)
  "Multiple keyclick handler.
Theoretically it can handle infinite keyclick sequence with self-recursive
call, but I guess emacs barfs perhaps around 150 clicks. Only SPC and
TAB are supported. See also `lkk-multiple-click-map' or `max-lisp-eval-depth'."
  (let* ((ev) (result) (continue)
	 (kdsc				; maybe passed from caller
	  (or kdsc (key-description (vector last-command-char)))))
    (or continue
	(catch 'hit
	  (mapcar
	   (lambda (pfx) 
	     (if (lookup-key
		  lkk-multiple-click-map 
		  (vector (make-symbol (concat pfx kdsc))))
		 (throw 'hit (setq continue t))))
	   (list "double-" "triple-" "quadruple-"))))
    (cond 
     (continue
      (cond 
       ((and (null (sit-for 0 lkk-multiple-click-threshold 'nodisp))
	     ;; This can't handle M-return (non-characer event) etc.
	     ;; (eq (setq ev (read-char-exclusive)) last-command-char)
	     (setq ev (read-event))
;;;          (setq eee (list kdsc ev (key-description (vector ev)))) ; debugger.
	     (null (listp ev))
	     (equal (key-description (vector ev)) kdsc)
	     )
	(setq hits (if hits (1+ hits) 2))
	;; Continue to call recursive.

	(lkk-multiple-click-crank hits kdsc)
	)
       ;; Else generate return value: caller *NEVER* drops into this.
       ;; Only one non-`last-command-char' key event is pushed back.
       ;; Normally unread-command-events seems to be empty so cons() is okay.
       (t
	;; MUST do ev push-back here on top only with the condition
	;; that ev != non-nil.  Don't know why.
	(if ev
	    (setq unread-command-events (cons ev unread-command-events)))
	(if (or (null hits) (> hits 4)) ; >4 shoud be caught by `catch' ?
	    nil
	  (lkk-multiple-click-execute hits kdsc)
	  )))))))

;; `while-map-over-on-list' method used here.
(defun lkk-get-command-key (keys)
  "Return key invoked `this-command' out of KEY removing prefix arg part."
  (cond ((fboundp 'this-single-command-keys) ; E19.34/20.5 has it. It's Nice.
	 (this-single-command-keys))
	(t 
	 (let ((command  this-command)
	       (klist (mapcar (lambda (c) c) keys))
	       (result))

	   (while klist
	     (if (eq command (key-binding (apply 'vector klist)))
		 (setq result (apply 'vector klist)))
	     (setq klist (cdr klist)))
	   result))))

(defun lkk-maneuver-old-keybind ()
  "The compatible function can be found in Good Old Classic skk as well."
  (let* ((skeys (lkk-get-command-key (this-command-keys)))
	 (cmd (let (lkk-j-mode
		    lkk-J-mode
		    lkk-l-mode
		    lkk-L-mode
		    lkk-q-mode
		    lkk-/-mode
		    lkk-mode)
		;; (setq lll (list skeys (key-binding skeys))) ;; debugger.
		;; Some non-character events should be trapped and
		;; replace with a character event (actually string
		;; here). Isn't it an emacs bug ?
		(cond ((and (vectorp skeys) (equal skeys [return]))
		       (setq skeys "\r")))
		(key-binding skeys)
		)))
    (if (and cmd (not (eq cmd this-command)))
	(call-interactively cmd) )))

;;; Old multiple-click handler supporting double-click-only
;;; (defsubst lkk-double-click-hack (hooks)
;;;   "Run HOOKS on double-click key hit.
;;; Double click status is defined by hitting another last-command-char
;;; within lkk-double-click-threshold.
;;; HOOKS are run via lkk-run-hooks, which returns the last hook value."
;;;   (let (ch)
;;;     ;; 'nodisp is better for this job ?
;;;     (if (null (sit-for 0 lkk-double-click-threshold 'nodisp))
;;; 	(if (eq (setq ch (read-char)) last-command-char)
;;; 	    ;; Eat double-click !!  ==> Return Double SPC ?
;;; 	    (lkk-run-hooks hooks)
;;; 	  ;; Push another event and return nil. Caller do the job.
;;; 	  (setq unread-command-events (cons ch unread-command-events))
;;; 	  ())
;;;       ;; ==> Return nil so that caller can do its job.
;;;       ())
;;;     ))

(defvar lkk-mclick-load-hook nil
  "A hook run at the end of loading `lkk-mclick'.")

(run-hooks 'lkk-mclick-load-hook)
