;;; lkk-skk-dichead.el --- Interface Method for SKK styled dictionary.

;;; At the bottom included lkk-dichead.el --- SKK -> lkk dict builder 

;;
;; GPL2 should be applied to functions and variables derived from skk or emacs.
;;
;; As for other code I wrote from scratch,
;; I just don't put any copying/copyright notice other than this:
;; Copyright (C) 2000 Seiichi Namba <sn@asahi-net.email.ne.jp>
;; There is of course no warranty.
;; Thu Apr 13 18:22:26 2000
;; s.n.
;;

;; The Dictionary Interface Layer should have only one entry point:
;; The function lkk-dict-search():
;;
;;   (defun lkk-dict-search (yomi &optional loose)
;;
;; which reads the dictionary if needed.
;; Currently lkk-dict-search is fset of lkk-skk-dict-search.

(provide 'lkk-skk-dichead)

;; Current lkk-skk-dict-search() in this file can handle okuri-nasi only.

;; Must do nothing unless suitable method is defined.
(cond ((eq lkk-dict-method 'lkk-skk-dict-buffer)
       (fset 'lkk-dict-search 'lkk-skk-dict-search)
       (fset
	'lkk-dict-update
	(lambda ()
	  (message
	   "lkk-dict-update not yet implemented in lkk-skk-dict-buffer method."
	    ) t)))
      (t nil))

;;;
;;; Kanji Conversion -- Low level dictionary file interface
;;; aka. Dictionary Manager -- should be split into discrete file ?
;;;

;;
;; The dark scanner
;;

;; New lkk-skk-dict-to-list()
;; which must be able to handle SKK entry such as:
;; / /$B!?(B/(concat "\057")/root/$B%k!<%H(B/
;;
;; (lkk-skk-dict-to-list str)
;; => ("/" "$B!?(B" "(concat \"\\057\")" "root" "$B%k!<%H(B")
;; (read (nth 2 (lkk-skk-dict-to-list str)))
;; => (concat "/")
;; (insert (eval (read (nth 2 (lkk-skk-dict-to-list str)))))
;; /
;; Nice !

(defsubst lkk-skk-dict-to-list (str)
  (let ((yomi) (rest) (delim "_~lkk~_"))
    ;; Yomi on BOL can include any string (eg. "/" itself).
    ;; So, take yomi and the rest part independently.
    (string-match "^[^ ]+ *" str)	; maybe * after SPC is not required...
    (setq yomi (substring str 0 (1- (match-end 0)))
	  ;; Remove `/' on top and tail or the `rest'.
	  rest (substring str (1+ (match-end 0)) -1))
    ;; Convert "/" to `delim' redundant enough, then split with the `delim'.
    (cons yomi (lkk-str-to-list (lkk-replace-in-string "/" delim rest) delim))
    ))

;; Old one. Can't handle spaces included in SKK entry's KANJI part.
;; (defsubst lkk-skk-dict-to-list-old (str)
;;   "Convert raw dict string to list"
;;   (lkk-str-to-list
;;    (substring (lkk-replace-in-string  " */" " " str) 0 -1)))

(defun lkk-skk-dict-get-current-record ()
  "Return current entry -- current as list."
  (lkk-skk-dict-to-list (lkk-get-current-line)))

;;
;; For okuri-ari entry hereunder.
;;

;;
;; Level 0 converter for `okuri-ari'
;;
;; Geneletes level 0 list.
;; (setq e "$B$J(Br /$B@.(B/$B47(B/$BLD(B/$BFk(B/[$B$k(B/$B@.(B/$B47(B/$BLD(B/]/[$B$l(B/$B47(B/$B@.(B/$BLD(B/$BFk(B/]/[$B$j(B/$B@.(B/$BLD(B/$B47(B/]/[$B$i(B/$BLD(B/$B47(B/$B@.(B/]/")
;; (lkk-skk-dict-to-list-okuri-0 e) =>
;; ("$B$J(Br" "$B@.(B/$B47(B/$BLD(B/$BFk(B/" "[$B$k(B/$B@.(B/$B47(B/$BLD(B/]" "[$B$l(B/$B47(B/$B@.(B/$BLD(B/$BFk(B/]" "[$B$j(B/$B@.(B/$BLD(B/$B47(B/]" "[$B$i(B/$BLD(B/$B47(B/$B@.(B/]")
;;
;; This sort of entry is found in my .skk-jisyo.  Is this correct ? 
;; $B$"$*(Bk /$B@D(B/
(defsubst lkk-skk-dict-to-list-okuri-0 (s)
  "Level 0 converter for converting dicstring to list \(for okuri-ari\)."
  (let (lis (regex0 " ")
	    (regex1 "\\[[^]]+\\]"))	; matches "[$B$k(B/$B@.(B/$B47(B/$BLD(B/]"
    (setq lis (cons (substring s 0 (string-match regex0 s)) lis)
	  s (substring s (1+ (match-end 0))))
    (cond
     ((string-match regex1 s)
      (setq lis (cons (substring s 0 (match-beginning 0)) lis)
	    s (substring s (match-beginning 0)))

      (while (string-match regex1 s)
	(setq lis (cons (substring s (match-beginning 0) (match-end 0)) lis)
	      s (substring s (1+ (match-end 0))))))
     (t (setq lis (cons s lis))))
    (nreverse lis)))

;;
;; Level 1 (entry point) converter for `okuri-ari'
;; 
;; Is this structure right ?
;; (lkk-skk-dict-to-list-okuri-1 en) =>
;; ("$B$J(Br"
;;  ("$B@.(B" "$B47(B" "$BLD(B" "$BFk(B")
;;  ("$B$k(B" "$B@.(B" "$B47(B" "$BLD(B")
;;  ("$B$l(B" "$B47(B" "$B@.(B" "$BLD(B" "$BFk(B")
;;  ("$B$j(B" "$B@.(B" "$BLD(B" "$B47(B")
;;  ("$B$i(B" "$BLD(B" "$B47(B" "$B@.(B"))
;;
(defsubst lkk-skk-dict-to-list-okuri (dicstr)
  "Level 1 converter for converting dicstring to list \(for okuri-ari\)."
  (let ((level0 (lkk-skk-dict-to-list-okuri-0 dicstr)))
    (cons
     (car level0)
     (mapcar
      (lambda (s)
	(cond
	 ((string= (substring s 0 1) "[")
	  (lkk-str-to-list
	   (substring
	    (lkk-replace-in-string "/" " " (substring s  1 -1)) 0 -1)))
	 (t
	  (lkk-str-to-list
	   (substring
	    (lkk-replace-in-string "/" " " s) 0 -1))))
	) (cdr level0)))))

(defun lkk-skk-dict-get-current-record-okuri ()
  "Return current entry -- current as list \(for okuri-ari\)."
  (lkk-skk-dict-to-list-okuri (lkk-get-current-line)))

;; Would-be alias of lkk-dict-search
;; lkk-dict-read-dict() sees lkk-okuri to determine search starting point.
(defun lkk-skk-dict-search (yomi &optional loose)
  "Returns dictionary entry lists or nil if failed.
Accepts YOMI, START (should be marker symbol), and LOOSE search flag."
  (save-excursion
    (set-buffer (lkk-dict-get-dict))	; in lkk.el
    ;; Currently, okuri is not supported in this skk-dichead.
    (goto-char (if lkk-okuri
		   lkk-okuri-ari-dict-top
		   lkk-okuri-nasi-dict-top))
    (cond ((search-forward (concat "\n" yomi (if loose "" " ")) (point-max) t)
;;; Using "^" slows down search very much.
;;;   600ms -> 2.1s for 300kb dict (search for last entry "$B$o$s$-$g$/(B")
;;;     (re-search-forward (concat "^" yomi (if strict " " "")) (point-max) t)
	   (forward-line 0)
	   (lkk-skk-dict-get-current-record))
	  (t nil))))
;;(lkk-skk-dict-read-dict) ;; Now it's done by lkk-dict-get-dict().

;;;;; Bare Test versions hereunder.

;;; Bare Test version. Conversion to 1st candidate only.
;;;(defun lkk-skk-dict-henkan-with-arg (arg)
;;;  (interactive "*P")
;;;  (if arg
;;;      (cond 
;;;       ((numberp arg)
;;;	(setq arg (prefix-numeric-value arg))
;;;	(let ((rb (save-excursion (backward-char arg) (point)))
;;;	      (re (point)))
;;;	  (lkk-convert-region rb re)))
;;;       ((listp arg)
;;;	(lkk-convert-region (region-beginning) (region-end)))
;;;       (t nil))
;;;    nil))

;;;(defun lkk-skk-dict-henkan (yomi &optional dic-index dic-entry)
;;;  "Return kanji for YOMI.
;;;Uses optional 2nd and 3rd arg DIC-ENTRY and DIC-INDEX
;;;if they are both non-nil."
;;;  (cond ((and dic-entry dic-index)
;;;      (if (string= yomi (lkk-get-yomi dic-entry))
;;;	  (nth dic-index dic-entry)))
;;;	(t
;;;	 (if (string= yomi (lkk-get-yomi lkk-current-record))
;;;	     (nth lkk-kanji-index lkk-current-record)))))

;;;
;;; lkk-dichead.el --- New skk2lkk dict converter
;;; Thu Feb 24 20:08:07 2000 s.n.
;;;

;; Perhaps requiring lkk is just for lkk-dict in `make lkkdict'.
;; On running lkk, lkk.el will explicitly load this program if needed.
;;(require 'lkk)

(or (boundp 'lkk-dict-header) (defconst "LKK dictionary")) ;; failsafe

;; Maybe LKKSKKDICT is passed by Makefile.
(defvar lkk-skk-dict-file (or (getenv "LKKSKKDICT") "~/.skk-jisyo"))

(defconst lkk-new-dict-name "./lkk-dict.new"
"lkk dictionary name to be generated in current directory.")

;;
;; Now this can handle
;; / /$B!?(B/(concat "\057")/root/$B%k!<%H(B/
;;  =>
;; ("/" "$B!?(B" "(concat "\\057")" "root" "$B%k!<%H(B")
;; sort of conversion.
;;

(defun lkk-skk2lkk (&optional skkdic lkkdic)
  "Read ~/.skk-jisyo, convert it into ./lkk-dict.new destructively.
SKKDIC, and LKKDIC filenames can be specified as optional 2nd and 3rd arg."
  (let* ((inhibit-quit t)
	 (skkdic (or skkdic lkk-skk-dict-file))
	 (skkbufname "**skk-jisho**")
	 (skkbuf (progn (if (get-buffer skkbufname)
			 (kill-buffer skkbufname))
		     (get-buffer-create skkbufname)))
	 (lkkbufname "*lkk-newdict*")
	 (lkkdic (or lkkdic lkk-new-dict-name))
	 (lkkbuf (progn (if (get-buffer lkkbufname)
			    (kill-buffer lkkbufname))
			(get-buffer-create lkkbufname)))
	 (okuri-nasi-dict-top)
	 tem)

    (save-excursion			; top header
      (set-buffer lkkbuf)

      ;;(message "lkkbuf: %s" buffer-file-coding-system) (sit-for 5)

      (goto-char (point-min))
      (insert (format ";; %s VERSION %s\n" lkk-dict-header lkk-version))
      (insert ";; okuri-ari\n"))

    (save-excursion
      (set-buffer skkbuf)
      (lkk-insert-file-contents skkdic)

      ;;(message "skkbuf: %s" buffer-file-coding-system) (sit-for 5)

      (setq okuri-nasi-dict-top
	    (save-excursion
	      (search-forward "\n;; okuri-nasi" (point-max) t)
	      (forward-line 1)
	      (point)))

      (widen)
      ;; okuri-ari
      (narrow-to-region (point-min) okuri-nasi-dict-top)
      (goto-char (point-min))
      (lkk-skk2lkk-subr 'lkk-skk-dict-get-current-record-okuri lkkbuf)

      (save-excursion			; 2nd header
	(set-buffer lkkbuf)
	(insert ";; okuri-nasi\n"))
      
      (widen)
      ;; okuri-nasi
      (narrow-to-region okuri-nasi-dict-top (point-max))
      (goto-char (point-min))
      (lkk-skk2lkk-subr 'lkk-skk-dict-get-current-record lkkbuf)
      )
    (save-excursion
      (set-buffer lkkbuf)
      (widen)
      (write-region (point-min) (point-max) lkkdic))))

(defun lkk-skk2lkk-subr (fun outbuf)
  ;; Used by lkk-skk2lkk-subr.
  (let ((newent) (count 0))
    (while (null (eobp))
      (if (looking-at "^;")
	  ()
	(setq newent (funcall fun))
	(save-excursion
	  (set-buffer outbuf)

	  ;; For REAL lisp expression version, use:

	  ;;(let ((standard-output (current-buffer)))
	  ;;  (prin1 newent) (insert ?\n))
	  (prin1 newent (current-buffer)) (insert ?\n)

;;; Try pp version. Fri Feb 25 14:32:09 2000
	  ;; This makes pretty printed okuri-ari, size is not big, but
	  ;; An entry is not in 1 line. Unused.
	  ;; (require 'pp) (pp newent (current-buffer))
;;; pp is slow:
;;; make lkkdict 2> /dev/null  27.86s user 0.41s system 98% cpu 28.563 total
;;; Above (prin1 newent (current-buffer)) (insert ?\n) case:
;;; make lkkdict 2> /dev/null  11.53s user 0.31s system 99% cpu 11.893 total

	  ;;(insert (format "%S\n" newent)) ; %S uses prin1. Seems like slower.

	  ;; (insert (format "%s\n" newent))

	  ;; Consideration on using format() on recording above:
	  ;; It makes `pseudo'
	  ;;   ($B$+$s$8(B $B4A;z(B $B46$8(B $B44;v(B)
	  ;;   ($B$J$,(Bs ($BD9(B $BN.(B) ($B$5(B $BD9(B $BN.(B) ($B$9(B $BN.(B $BD9(B) ($B$7(B $BN.(B $BD9(B) ($B$=(B $BN.(B))
	  ;; sort of expression -- can't be read()/read-from-string() directly,
	  ;; because they lack string quotation "", but fairly small
	  ;; (even smaller than original .skk-jisyo) lkk-dict.  Perhaps direct
	  ;; editing on dict buffer is easier than real lisp
	  ;; expression such as
	  ;; ("$B$+$s$8(B" "$B4A;z(B" "$B46$8(B" "$B44;v(B")

	  ;; BUT not taken as LKK standard. Reason:
	  ;; Can't store this kind of 
	  ;; / /$B!?(B/(concat "\057")/root/$B%k!<%H(B/
	  ;;  skk=>lkk
	  ;; ("/" "$B!?(B" "(concat "\\057")" "root" "$B%k!<%H(B")
	  ;; entry properly.  This sort of conversion can't be handled with
	  ;; current lkk-skk2lkk(), though.

	  ))

      ;; Echo (to stderr if -batch) count and new entries.
      (message "%06d: %s" (setq count (1+ count)) (prin1-to-string newent))
      ;; This is not so fast. Show real one by prin1. (princ
      ;;(message "%06d: %s" (setq count (1+ count)) newent)

      (forward-line 1))))

;; Convert pseudo-lisp expression generated by format() into real one.
;; A sort of prin1 compatible. Found to be unnecessary (use prin1).
;; (insert (lkk-gen-lisp-string '("a" bb (cc "d") ("e" "f") )))
;; ("a" 'bb ('cc "d") ("e" "f"))
;; (defun lkk-gen-lisp-string (list)
;;   "Read LIST and return Readable list. Supports only string and sublist."
;;   (format
;;    "(%s)"
;;    (let ((outstr "") c)
;;      (while (setq c (car list))
;;        (cond
;; 	((listp c)
;; 	 (setq outstr
;; 	       (concat outstr (lkk-format-to-expression c)
;; 		       (if (cdr list) " " ""))))
;; 	((stringp c)
;; 	 (setq outstr
;; 	       (concat outstr "\"" c
;; 		       (if (cdr list) "\" " "\""))) )
;; 	(t				; Perhaps symbol. Do not support vector.
;; 	 (setq outstr
;; 	       (concat outstr (format "'%s" c) (if (cdr list) " " ""))) ))
;;        (setq list (cdr list)))
;;      outstr)))

;; Don't do it right now. Use `-f' from `-batch'
;; (lkk-skk2lkk)
